''
'' VGA display 50x18 (single cog, ROM font) - video driver and pixel generator
''
''        Author: Marko Lukat
'' Last modified: 2013/10/03
''       Version: 0.2
''
'' long[par][0]:  screen: [!Z]:addr = 16:16 -> zero (accepted), 2n
'' long[par][2]: palette: [!Z]:addr = 16:16 -> zero (accepted), 4n
'' long[par][3]: frame indicator
''
'' 20131003: initial version (800x600@60Hz timing, %00 sync locked)
''
VAR
  byte cogNum

PUB start(dataptr)
  return cogNum := cognew(@driver, dataptr)       ' start the PASM overlay driver

PUB stop
  if cogNum
    cogstop(cogNum)
  cogNum~
  
DAT           org     0                       ' video driver

driver        mov       graphptr, par              ' because the only mailbox parameter IS "graph."  Not planning on making a super-configurable routine...yet!
              'NOT giving options for "configurable Y position, configurable X-range", etc.  Hardcode for one purpose.  That greatly simplifies the init routine.

              neg       href, cnt               '  -4   hub window reference            (%%)      ??????????????????????
        
' Upset video h/w and relatives.

                movi    ctrb, #%0_11111_000     ' LOGIC always (loader support)                         '###?????  VCO+128 (lower three), counter mode "ALWAYS", no output pin (upper five)

                movi    ctra, #%0_00001_110     ' PLL, VCO/2
                movi    frqa, #%0001_00000      ' 5MHz * 16 / 2 = 40MHz = pixel clock
                movs    frqa, #res_x/4          ' |
                movs    frqa, #0                ' insert res_x into phsa
                
                mov     vscl, hvis              ' 1/16
                mov     vcfg, vcfg_norm         ' VGA, 4 colour mode            ### was vcfg_sync

                rdlong  cnt, #0
                shr     cnt, #10                ' ~1ms
                add     cnt, cnt               
                waitcnt cnt, #0                 ' PLL needs to settle

                waitvid zero, #0                ' dummy (first one is unpredictable)
                waitvid zero, #0                ' point of reference

                add     href, cnt               ' get current sync slot
                shr     href, #1                ' 2 system clocks per pixel
                sub     href, #9                ' 9..16 >> 0..7
                neg     href, href              ' |
                and     href, #%111             ' calculate adjustment

' WHOP is reasonably far away so we can update vscl without re-sync.

                add     vscl, href              ' |
                waitvid zero, #0                ' stretch frame
                sub     vscl, href              ' |
                waitvid zero, #0                ' restore frame

' At this point all WHOPs are aligned to 16h+11. In fact they are aligned to 32h+11
' due to the frame counter covering 2 hub windows.
'Finished init of video generators, etc.

'FULL DISCLAIMER: I dunno what all that code above does...I'm just leaving it right there.
              mov     dira, mask              ' drive outputs.  NOTE: Does not drive VSYNC/HSYNC pins.

' horizontal timing 800(800) 40(40) 128(128) 88(88)
'   vertical timing 600(600)  1(1)    4(4)   23(23)
'=================================================================================================
vsync         

              ' An experiment here...
              '...just wait for the VSync signal from the video driver.
              ' We don't need to do the video sync/drive portion of anything.  Just overlaying on top of the already-generated signal.
'              waitcnt   cnt, #0                 ' re-sync for steady video.  RESULTS IN MESSED UP TEXT COLORS/NO GRAPH OUTPUT
         '     waitvid   sync, #0

              waitpeq   vsync_pin, vsync_pin    ' Wait for the VSync pin to go HIGH
              waitpne   vsync_pin, vsync_pin    ' Wait for the VSync pin to go LOW
              mov       yline, #0               ' starting at the top of the screen.
              mov       gLines, #graph_height   ' number of lines in the graph

' Synced to the VSync signal.  We need to wait a number of HSync lines to get out of the vertical blanking,
'and to where we want to output video data.
'------------------------------------------------------------------------------------------------
line          waitpeq   hsync_pin, hsync_pin    ' Wait for the HSync pin to go HIGH
              waitpne   hsync_pin, hsync_pin    ' Wait for the HSync pin to go LOW
              add       yLine, #1
              
              'Start of range: just pick line 300
              cmp       yLine, #graph_start    wc, wz
        if_c  jmp       #line
        if_z  mov       ycomp, ycompStart

              ' Why not write a little grid (using color #3)???
              test      ycomp, #$1F            wz
        if_z  movs      one, #grid
        if_nz movs      one, #allzero
              '----------------------------------
              'At or beyond #graph_start
              'Need to wait out past the blanking portion...
              mov       graphptr, par           'graph base address
              
              waitvid   sync, #0
              waitvid   sync, #0
              waitvid   sync, #0

              mov       xBlocks, #25            '800 / 16 = 50
              'read a byte.  Have 6 instructions available in this inner loop between waitvids...
pixel         rdword    temp, graphptr
              add       graphptr, #2

              waitvid   graph_colors, data
              sub       temp, ycomp             'can very easily mask for "zero" in each byte to see if we have a match.
one           mov       data, zero

              test      temp, mask_ch2    wz
        if_z  or        data, ch_2
              test      temp, #$FE    wz
        if_z  or        data, ch_1
              
              waitvid   graph_colors, data
              djnz      xBlocks, #pixel
              '----------------------------------

              waitvid   graph_colors, data
              waitvid   graph_colors, #$00      ' mute output
              'How many rows to display?
              sub       ycomp, #$0101           ' ...we can very easily check for overflow this way!
              djnz      gLines, #line
              'Done.  Wait for the next frame.
              jmp       #vsync                  ' next frame
'=================================================================================================

' initialised data and/or presets

sync          long    $0200                   ' locked to %00 {%hv}

hvis          long      1 << 12 | 16          '   1/16

vcfg_norm     long    %0_01_1_00_000 << 23 | pgrp << 9 | vpin
vsync_pin     long %01 << (pgrp * 8)
hsync_pin     long %10 << (pgrp * 8)

mask          long    vpin << (pgrp * 8) '### | %11 << (sgrp * 8) ###...removed VSync/HSync pins.  We're only trying to sync to the text screen 

graph_colors  long $3008C000                    ' R/G/B/black...

ch_1          long $55555555
ch_2          long $FFFFFFFF
mask_ch2      long $0000FE00
grid          long $22222222
allzero       long $00000000
ycompStart    long $0000FDFD
                
href          res       1                       ' hub window reference  < setup +3      (%%)
temp          res       1
graphptr      res       1                       ' screen buffer         < setup +4      (%%)
data          res       1
yline         res       1
ycomp         res       1       'value y-offset corrected for graph offset
gLines        res       1
xBlocks       res       1

tail          fit

CON
  zero    = $1F0                                ' par (dst only)
  vpin    = $0FC                                ' pin group mask
  pgrp    = 1                                   ' pin group
  dcolour = %%0220_0010                         ' default colour
  
  res_x   = 800                                 ' |
  res_y   = 600                                 ' |
  res_m   = 4                                   ' UI support

  graph_start = 320
  graph_height = 256