''
'' VGA display 50x18 (single cog, ROM font) - video driver and pixel generator
''
''        Author: Marko Lukat
'' Last modified: 2013/10/03
''       Version: 0.2
''
'' long[par][0]:  screen: [!Z]:addr = 16:16 -> zero (accepted), 2n
'' long[par][2]: palette: [!Z]:addr = 16:16 -> zero (accepted), 4n
'' long[par][3]: frame indicator
''
'' 20131003: initial version (800x600@60Hz timing, %00 sync locked)
''
PUB init(mailbox)
  return cognew(@driver, mailbox)       ' start the PASM video driver
  
DAT             org     0                       ' video driver

driver          jmpret  $, #setup               '  -4   once

                mov     dira, mask              ' drive outputs

' horizontal timing 800(800) 40(40) 128(128) 88(88)
'   vertical timing 600(600)  1(1)    4(4)   23(23)

vsync           mov     eins, scrn              ' screen base address
                mov     vier, plte              ' colour buffer
                mov     drei, vres              ' max visible scanlines

                mov     scnt, #27               ' scnt controls fetch (27..1)

'               mov     ecnt, #1
                call    #blank                  ' front porch (scnt == 27, inactive)
'               djnz    ecnt, #$-1

' Start fetch during last 27 blank lines.

                xor     sync, #$0101            ' active

                mov     ecnt, #4
                call    #blank                  ' vertical sync
                sub     scnt, #1                ' advance now
                djnz    ecnt, #$-2

                xor     sync, #$0101            ' inactive

                call    #blank                  ' back porch
                djnz    scnt, #$-1

' Vertical sync chain done, do visible area.

:line           mov     frqb, #0 wc             ' initial font row index
                mov     scnt, #32               ' font size
                max     scnt, drei              ' limit against what's left
                sub     drei, scnt              ' update what's left

:scan           waitcnt cnt, #0                 ' re-sync after back porch              (##)

                mov     vcfg, vcfg_norm         ' disconnect sync from video h/w        (&&)

                jmpret  hsync_ret, #emit        ' display scanline

                add     frqb, #2                ' next row in character definition(s)
                djnz    scnt, #:scan            ' repeat for font size
                tjnz    drei, #:line            ' repeat for all rows

                wrlong  cnt, fcnt_              ' announce vertical blank
                
                jmp     #vsync                  ' next frame


blank           mov     vscl, phsa              ' 256/800
                waitvid sync, #%%0000           ' latch blank line

hsync           mov     vscl, #40               ' 256/40
                waitvid sync, #%%0

                mov     vcfg, vcfg_sync         ' switch back to sync mode              (&&)

                mov     cnt, cnt                ' record sync point                     (##)
                add     cnt, #9{14}+436           

                mov     vscl, wrap              ' 128/216
                waitvid sync, #%%01

                call    #fetch
hsync_ret
blank_ret       ret


emit            mov     vscl, hvis              ' 1/16
                mov     ecnt, #50               ' character count

        if_nc   xor     $+2, #addr^(addr+50)    ' address array |
        if_nc   xor     $+2, #cols^(cols+50)    ' colour array  | primary <> secondary

                movs    :one, #addr             ' |
                movd    :two, #cols             ' reset
                
:loop           add     :one, #1                ' next address (delayed, need fast exit path)
                
:one            mov     phsb, addr              ' character pixmap address
                rdlong  temp, phsb              ' read from address + 2frqb (row adjustment)
:two            cmp     cols, temp              ' WHOP

                add     :two, dst1              ' next colour entry
                djnz    ecnt, #:loop            ' next 16px
    
emit_ret        jmpret  zero, #hsync wc,nr      ' set carry


fetch           cmp     drei, #0 wz             ' enabled?
        if_e    jmp     fetch_ret

                cmp     scnt, #26 wz,wc         ' 26..32
        if_ae   jmp     #fetch_26               ' re-initialise sequence

                call    #ascii
                call    #ascii
                
fetch_ret       jmpret  zero, #0-0 wc,nr        ' set carry


fetch_26        cmp     scnt, #26 wz            ' only once in 26..32
        
        if_e    xor     $+2, #addr^(addr+50)    ' |
        if_e    xor     $+2, #cols^(cols+50)    ' swap buffers

        if_e    movd    f_0, #addr              ' |
        if_e    movd    f_1, #cols              ' preset filler code
                jmpret  zero, fetch_ret wc,nr   ' set carry


ascii           rdbyte  temp, eins              ' fetch character
                add     eins, #1                ' advance address

                shr     temp, #1 wc             ' even/odd
                or      temp, #%1_00000000      ' ROM base address
                shl     temp, #7                ' final location

f_0             mov     0-0, temp
                add     $-1, dst1

                rdword  zwei, vier              ' fetch colour
                add     vier, #2                ' advance address

                mov     temp, zwei              ' $0000BBAA     
        if_c    shr     temp, #8
                shl     temp, #16               ' $BBAA0000/$00BB0000
                or      temp, zwei              ' $BBAABBAA/$00BBBBAA

        if_c    shl     temp, #8                ' $BBBBAA00
        if_c    and     zwei, #$FF
        if_c    or      temp, zwei              ' $BBBBAAAA
                
f_1             mov     0-0, temp
                add     $-1, dst1
                
ascii_ret       ret

' initialised data and/or presets

sync            long    $0200                   ' locked to %00 {%hv}

wrap            long    128 << 12 | 216         ' 128/216
hvis            long      1 << 12 | 16          '   1/16

vcfg_norm       long    %0_01_1_00_000 << 23 | vgrp << 9 | vpin
vcfg_sync       long    %0_01_1_00_000 << 23 | sgrp << 9 | %11

mask            long    vpin << (vgrp * 8) | %11 << (sgrp * 8)

scrn_           long    +0                      ' |
plte_           long    +8                      ' |
fcnt_           long    12                      ' mailbox addresses (local copy)

vres            long    res_y
dst1            long    1 << 9                  ' dst +/-= 1
                
' Stuff below is re-purposed for temporary storage.

setup           add     scrn_, par              '  +0 = @long[par][0]
                add     plte_, par              '  +4   @long[par][2]
                add     fcnt_, par              '  +8   @long[par][3]

                neg     href, cnt               '  -4   hub window reference            (%%)

                rdlong  scrn, scrn_ wz          ' get screen address (2n)               (%%)
        if_nz   wrlong  zero, scrn_             ' acknowledge screen buffer setup

                rdlong  plte, plte_ wz          ' get font colour                       (%%)
        if_nz   wrlong  zero, plte_             ' acknowledge font colour
        
' Upset video h/w and relatives.

                movi    ctrb, #%0_11111_000     ' LOGIC always (loader support)

                movi    ctra, #%0_00001_110     ' PLL, VCO/2
                movi    frqa, #%0001_00000      ' 5MHz * 16 / 2 = 40MHz
                movs    frqa, #res_x/4          ' |
                movs    frqa, #0                ' insert res_x into phsa
                
                mov     vscl, hvis              ' 1/16
                mov     vcfg, vcfg_sync         ' VGA, 4 colour mode

                rdlong  cnt, #0
                shr     cnt, #10                ' ~1ms
                add     cnt, cnt               
                waitcnt cnt, #0                 ' PLL needs to settle

                waitvid zero, #0                ' dummy (first one is unpredictable)
                waitvid zero, #0                ' point of reference

                add     href, cnt               ' get current sync slot
                shr     href, #1                ' 2 system clocks per pixel
                sub     href, #9                ' 9..16 >> 0..7
                neg     href, href              ' |
                and     href, #%111             ' calculate adjustment

' WHOP is reasonably far away so we can update vscl without re-sync.

                add     vscl, href              ' |
                waitvid zero, #0                ' stretch frame
                sub     vscl, href              ' |
                waitvid zero, #0                ' restore frame

' At this point all WHOPs are aligned to 16h+11. In fact they are aligned to 32h+11
' due to the frame counter covering 2 hub windows.

                jmp     %%0                     ' return

                fit
                
' uninitialised data and/or temporaries

                org     setup

scnt            res     alias                   ' scanlines (per char)
href            res     1                       ' hub window reference  < setup +3      (%%)
scrn            res     1                       ' screen buffer         < setup +4      (%%)
plte            res     1                       ' colour buffer         < setup +6      (%%)
ecnt            res     1                       ' element count
temp            res     1

eins            res     1
zwei            res     1
drei            res     1
vier            res     1

addr            res     50 *2
cols            res     50 *2

tail            fit
                
DAT                                             ' translation table

__table         word    (@__names - @__table)/2

                word    res_x
                word    res_y
                word    res_m
                
__names         byte    "res_x", 0
                byte    "res_y", 0
                byte    "res_m", 0

CON
  zero    = $1F0                                ' par (dst only)
  vpin    = $0FC                                ' pin group mask
  vgrp    = 1                                   ' pin group
  sgrp    = 1                                   ' pin group sync
  dcolour = %%0220_0010                         ' default colour
  
  res_x   = 800                                 ' |
  res_y   = 600                                 ' |
  res_m   = 4                                   ' UI support

  alias   = 0
  
DAT